#include <stdlib.h>
#include <stdio.h>
#include <cuda.h>
#include <curand.h>

class __vector {
  public:
	double x, y, z;

	__device__ __vector();
	__device__ __vector(double x, double y, double z);

	__device__ double dot(__vector w);
	__device__ double operator*(__vector w);
	__device__ __vector cross(__vector w);
	
	__device__ __vector add(__vector w);
	__device__ __vector operator+(__vector w);
	__device__ __vector sub(__vector w);
	__device__ __vector operator-(__vector w);
	__device__ __vector mult(double s);
	__device__ __vector operator*(double s);
	
	__device__ __vector h(__vector w);
	__device__ double length();
	__device__ __vector unit();
};

__device__ __vector::__vector() : x(0.0), y(0.0), z(0.0) { }
__device__ __vector::__vector(double x, double y, double z) : x(x), y(y), z(z) { }

__device__ double __vector::dot(__vector w) { return this->x*w.x + this->y*w.y + this->z*w.z; }
__device__ double __vector::operator*(__vector w) { return this->x*w.x + this->y*w.y + this->z*w.z; }
__device__ __vector __vector::cross(__vector w) { return __vector(this->y * w.z - this->z * w.y, this->z * w.x - this->x * w.z, this->x * w.y - this->y * w.x); }

__device__ __vector __vector::add(__vector w) { return __vector(this->x + w.x, this->y + w.y, this->z + w.z); }
__device__ __vector __vector::operator+(__vector w) { return __vector(this->x + w.x, this->y + w.y, this->z + w.z); }
__device__ __vector __vector::sub(__vector w) { return __vector(this->x - w.x, this->y - w.y, this->z - w.z); }
__device__ __vector __vector::operator-(__vector w) { return __vector(this->x - w.x, this->y - w.y, this->z - w.z); }
__device__ __vector __vector::mult(double s) { return __vector(s * this->x, s * this->y, s * this->z); }
__device__ __vector __vector::operator*(double s) { return __vector(s * this->x, s * this->y, s * this->z); }

__device__ __vector __vector::h(__vector w) { return __vector(this->x * w.x, this->y * w.y, this->z * w.z); }
__device__ double __vector::length() { return sqrt(this->x*this->x + this->y*this->y + this->z*this->z); }
__device__ __vector __vector::unit() {
	double mag = sqrt(this->x*this->x + this->y*this->y + this->z*this->z);
	return __vector(this->x/mag, this->y/mag, this->z/mag);
}

struct __ray {
	__vector origin;
	__vector direction;
};

struct __intersection {
	bool intersects;
	double t;
	
	__ray ray;
	__vector normal;
};

enum geometry  { NONE, PLANE, SPHERE };
enum materials { DIFFUSE, SPECULAR, REFRACTIVE };

struct __object {
	int type; // geometry
	int material; // material
	double radius;

	__vector emission;
	__vector color;
	__vector normal;
	__vector point;
	__vector center;
};





__device__ __vector sampleRay(__ray ray, __object objects[], int object_count, int depth, float* rand_device, int index, int size, int max_bounces) {
	double epsilon = 0.000001;

	__intersection intersect;
	__intersection isect;
	
	int which = -1;
	int num_objects = object_count;

	int m = max_bounces > 10 ? 10 : max_bounces;
	__vector __a[10], __b[10], sample(0,0,0);

	for (int l = 0; l < m; l++) {
		__a[l].x = __a[l].y = __a[l].z = 0;
		__b[l].x = __b[l].y = __b[l].z = 0;
	}

	for (int l = 0; l < m; l++) {

		ray.direction = ray.direction.unit();

		intersect.intersects = false;

		for (int k = 0; k < num_objects; k++) {
			isect.intersects = false;

			if (objects[k].type == SPHERE) {

				double a = ray.direction * ray.direction;
				double b = (ray.direction * ray.origin - ray.direction * objects[k].center) * 2.0;
				double c = ray.origin * ray.origin + objects[k].center * objects[k].center - ray.origin * objects[k].center * 2.0 - objects[k].radius * objects[k].radius;
				double det = b * b - 4 * a * c;
				if (det < epsilon) continue;

				double t0 = (-b + sqrt(det))/(2 * a);
				double t1 = (-b - sqrt(det))/(2 * a);
				if (t0 < epsilon && t1 < epsilon) continue;

				isect.intersects = true;
				isect.t = t0 < epsilon ? t1 : (t1 < epsilon ? t0 : (t0 < t1 ? t0 : t1));

				isect.ray.origin = ray.origin + ray.direction * isect.t;

				isect.normal = (isect.ray.origin - objects[k].center).unit();

			} else if (objects[k].type == PLANE) {

				double den = ray.direction * objects[k].normal;
				if (fabs(den) < epsilon) continue;

				__vector temp = objects[k].point - ray.origin;
				double num = temp * objects[k].normal;
				double num_den = num/den;
				if (num_den < epsilon) continue;

				isect.intersects = true;
				isect.t = num_den;

				isect.ray.origin = ray.origin + ray.direction * isect.t;

				isect.normal = objects[k].normal.unit();

			}

			if (isect.intersects) {
				if (!intersect.intersects || isect.t < intersect.t) {
					intersect = isect;
					which = k;
				}
			}
		}

		if (intersect.intersects) {

			__a[l] = objects[which].emission;
			__b[l] = objects[which].color;
			
			//if (which == 3 || which == 8) break; // light or black plane hack

			// reflection for new ray
			intersect.ray.direction = (ray.direction - intersect.normal * (ray.direction * intersect.normal * 2.0)).unit();

			if (objects[which].material == DIFFUSE) { //DIFFUSE

				__vector w = intersect.normal;

				// cosine weighted sampling
//				double u1 = rand_device[index+0];
//				double u2 = rand_device[index+1];
				double u1 = rand_device[size*(l+1) + index+0];
				double u2 = rand_device[size*(l+1) + index+1];

				double r1 = 2 * M_PI * u1;
				double r2 = sqrt(1 - u2);
				double r3 = sqrt(u2);

				__vector u(0,0,0);
				if      (fabs(w.x) < fabs(w.y) && fabs(w.x) < fabs(w.z)) u.x = 1;
				else if (fabs(w.y) < fabs(w.x) && fabs(w.y) < fabs(w.z)) u.y = 1;
				else u.z = 1;

				u = u.cross(w).unit();
				__vector v = w.cross(u).unit();
					 u = v.cross(w).unit();
				__vector d = ( u * (cos(r1) * r2) + v * (sin(r1) * r2) + w * r3 ) .unit();

				intersect.ray.direction = d;

				ray = intersect.ray;
				
			} else if (objects[which].material == SPECULAR) { //SPECULAR

				ray = intersect.ray;
				
			} else if (objects[which].material == REFRACTIVE) { //REFRACTIVE

				bool into = ray.direction * intersect.normal < 0; // entering the medium

				double n1n2 = into ? (1.0/1.5) : (1.5/1.0);
				__vector n  = into ? intersect.normal : (intersect.normal * -1);
				__vector r = ray.direction;
				
				double n1n22 = n1n2 * n1n2;
				double rn   = r * n;
				double rn2  = rn * rn;
				
				double a = 1 - n1n22 * (1 - rn2);
				if (a >= 0) {
					ray.origin = intersect.ray.origin;
					ray.direction = r * n1n2 - n * (n1n2 * rn + sqrt(a));
				} else ray = intersect.ray; // total internal reflection

			}
			
		} else break;
	}

	sample = __a[m - 1];
	for (int l = m-2; l >= 0; l--) sample = __a[l] + __b[l].h(sample);

	return sample;
}

__global__ void kernel(unsigned char* current_buffer_device, double* current_doubles_device, float* rand_device, __object objects[], int samples, int object_count, int max_bounces, int width, int height, double focal_distance, double blur_radius) {
	int x = threadIdx.x + blockIdx.x * blockDim.x;
	int y = threadIdx.y + blockIdx.y * blockDim.y;

	int index = y * (width * 3) + x * 3;

	__vector _samples;

	for (int i = 0; i < 4; i++) {
		// cosine weighted sampling
		double u1 = rand_device[i*width*height*3+index+0];
		double u2 = rand_device[i*width*height*3+index+1];
		double r1 = 2 * M_PI * u1;
		double r2 = sqrt(1 - u2);
		double r3 = sqrt(u2);
		__vector offset = __vector(cos(r1)*r2, sin(r1)*r2, r3) * 0.5;
/*/
		double u1 = rand_device[(i*4)*width*height*3+index+0] * 2.0;
		double u2 = rand_device[(i*4)*width*height*3+index+1] * 2.0;
		double dx = u1 < 1 ? (sqrt(u1) - 1) : (1- sqrt(2-u1));
		double dy = u2 < 1 ? (sqrt(u2) - 1) : (1- sqrt(2-u2));
		dx /= 2;
		dy /= 2;
		__vector offset = __vector(dx, dy, 0);
*/
		if (i == 0) offset = offset + __vector( 0.0, 0.0, 0.0);
		if (i == 1) offset = offset + __vector( 0.0, 0.5, 0.0);
		if (i == 2) offset = offset + __vector( 0.5, 0.0, 0.0);
		if (i == 3) offset = offset + __vector( 0.5, 0.5, 0.0);

		__vector dir = __vector(x - width  / 2, -y + height / 2, 0 + width) + offset;
		__ray ray = { __vector(0, 0, 0), dir.unit() };

//		__vector sample = sampleRay(ray, objects, object_count, 0, rand_device, index+width*height*3*((i+1)*4), width*height*3, max_bounces);

//		_samples = _samples + sample * 0.25;

// apply depth of field
		// http://mzshehzanayub.blogspot.com/2012/10/gpu-path-tracer.html
		u1 = rand_device[i*width*height*3+index+1];
		u2 = rand_device[i*width*height*3+index+2];
		r1 = 2 * M_PI * u1;
		r2 = u2;
		//r3 = sqrt(u2);
		offset = __vector(cos(r1)*r2, sin(r1)*r2, 0.0) * blur_radius;//.unit();
//		__vector p = ray.origin + ray.direction * f;
		__vector p = ray.origin + dir * (focal_distance / width);
		ray.origin = ray.origin + offset;
		ray.direction = (p - ray.origin).unit();

		__vector sample = sampleRay(ray, objects, object_count, 0, rand_device, index+width*height*3*((i+1)*4), width*height*3, max_bounces);

		_samples = _samples + sample * 0.25;
	}
	
	// add the sample to the  accumulation
	current_doubles_device[index+0] = (current_doubles_device[index + 0] * (samples - 1.0) + _samples.x) / samples;
	current_doubles_device[index+1] = (current_doubles_device[index + 1] * (samples - 1.0) + _samples.y) / samples;
	current_doubles_device[index+2] = (current_doubles_device[index + 2] * (samples - 1.0) + _samples.z) / samples;

	// save the current frame
	current_buffer_device[index+0] = (current_doubles_device[index+0] * 255) > 255 ? 255 : (int)(current_doubles_device[index+0] * 255);
	current_buffer_device[index+1] = (current_doubles_device[index+1] * 255) > 255 ? 255 : (int)(current_doubles_device[index+1] * 255);
	current_buffer_device[index+2] = (current_doubles_device[index+2] * 255) > 255 ? 255 : (int)(current_doubles_device[index+2] * 255);
}

bool initializePathTracer(unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, __object* _objects, __object*& current_objects_device, int object_count, float*& rand_device, int max_bounces, int width, int height) {
	int num_bytes;
	
	// for storing the image result
	num_bytes = sizeof(unsigned char) * width * height * 3;
	current_buffer_host = (unsigned char*)malloc(num_bytes);
	cudaMalloc((void**)&(current_buffer_device), num_bytes);

	// for accumulating samples
	num_bytes = sizeof(double) * width * height * 3;
	cudaMalloc((void**)&(current_doubles_device), num_bytes);
	cudaMemset(current_doubles_device, 0, num_bytes);
	
	// for generating random uniforms
	int m = 4 + max_bounces * 4;
	num_bytes = sizeof(float) * width * height * 3 * m;
	cudaMalloc((void**)&rand_device, num_bytes);
	
	// for storing the objects on the device
	num_bytes = sizeof(__object) * object_count;
	cudaMalloc((void**)&current_objects_device, num_bytes);
	cudaMemcpy(current_objects_device, _objects, num_bytes, cudaMemcpyHostToDevice);

	return true;
}

bool setupPathTracer(double*& current_doubles_device, __object*& current_objects_device, __object* _objects, int object_count, int width, int height) {
	int num_bytes = sizeof(double) * width * height * 3;

	// zero out the sample accumulation
	cudaMemset(current_doubles_device, 0, num_bytes);
	
	// store the objects on the device
	num_bytes = sizeof(__object) * object_count;
	cudaMemcpy(current_objects_device, _objects, num_bytes, cudaMemcpyHostToDevice);
	
	return true;
}

bool releasePathTracer(unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, __object*& current_objects_device, float*& rand_device, int width, int height) {

	free(current_buffer_host);
	cudaFree(current_buffer_device);

	cudaFree(current_doubles_device);

	cudaFree(rand_device);

	cudaFree(current_objects_device);
	
	return true;
}

int runPathTracer(unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, float*& rand_device, unsigned long long samples, __object objects[], __object*& current_objects_device, int object_count, int max_bounces, int width, int height, double focal_distance, double blur_radius) {

	int num_bytes;
	int dimx = 16;
	int dimy = 16;
	dim3 dimGrid(width/dimx, height/dimy);
	dim3 dimBlock(dimx, dimy);

	// generate random uniforms
	int m = 4 + max_bounces * 4; // 1 plus 10 levels
	curandGenerator_t gen;
	curandCreateGenerator(&gen, CURAND_RNG_PSEUDO_MTGP32);
	curandSetPseudoRandomGeneratorSeed(gen, samples); //samples=seed
	curandGenerateUniform(gen, rand_device, width * height * 3 * m);
	curandDestroyGenerator(gen);

	//cudaDeviceSetLimit (cudaLimitStackSize, );
	//cudaThreadSetLimit (cudaLimitStackSize, 8192*16)
	

	kernel<<<dimGrid, dimBlock>>>(current_buffer_device, current_doubles_device, rand_device, current_objects_device, samples, object_count, max_bounces, width, height, focal_distance, blur_radius);


	num_bytes = sizeof(unsigned char) * width * height * 3;
	printf("%s\n", cudaGetErrorString(
		cudaMemcpy(current_buffer_host, current_buffer_device, num_bytes, cudaMemcpyDeviceToHost)
	));

	return samples;
}